/*  -*- Mode: C -*-  */

/* regress-test.c --- miscellaneous tests for snprintfv
 * Copyright (C) 2002 Gary V. Vaughan
 * Originally by Paolo Bonzini, 2002
 * This file is part of Snprintfv.
 *
 * Snprintfv is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * Snprintfv is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * As a special exception to the GNU General Public License, if you
 * distribute this file as part of a program that also links with and
 * uses the libopts library from AutoGen, you may include it under
 * the same distribution terms used by the libopts library.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <float.h>
#include <stdio.h>
#include <stdarg.h>
#include "config.h"
#include "printf.h"

static int test = 0;
static int wrong = 0;
void regress (const char *result, const char *ok)
{
  ++test;
  if (strcmp (result, ok))
    {
      ++wrong;
      printf ("===========================================\n"
              "Test %d failed!\n"
              "===========================================\n"
              "Result\n"
              "------\n"
              "%s\n"
              "\n"
              "Correct\n-------\n"
              "%s\n\n", test, result, ok);
    }
}

void test1 ()
{
  static const char zDne[] = "%6$s"
    "%1$sDO NOT EDIT THIS FILE   (%2$s)\n"
    "%1$s\n"
    "%1$sIt has been AutoGen-ed  %3$s\n"
    "%1$sFrom the definitions    %4$s\n"
    "%1$sand the template file   %5$s";

  static const char zDneOk[] = "six "
    "one DO NOT EDIT THIS FILE   (two )\n"
    "one \n"
    "one It has been AutoGen-ed  three \n"
    "one From the definitions    four \n"
    "one and the template file   five ";

  char *x;
  asprintf(&x, zDne, "one ", "two ", "three ", "four ", "five ", "six " );
  regress (x, zDneOk);
}

void
test2 ()
{
#define bufSize 0x4000
  char zBuf[ bufSize ];
  char *args[3];

  static const char zFmt[] =
    "%s copyright (c) %s %s, all rights reserved\n";

  static const char zFmtOk[] =
    "autogen copyright (c) 1992-2002 Bruce Korb, all rights reserved\n";

  args[0] = "autogen";
  args[1] = "1992-2002";
  args[2] = "Bruce Korb";
 
  snprintfv( zBuf, bufSize, zFmt, args );
  regress (zBuf, zFmtOk);
}

#define STRING(x) STRING_(x)
#define STRING_(x) #x

/* Prevent some optimization attempts from the compiler */
void 
x ()
{
}

void
test3_5 ()
{
  float f,g,h;
  float one = 1.0;
  float two = 2.0;
  float four = 4.0;
  char zBuf[ bufSize ];

  /* Test the machine epsilon.

     For the machine epsilon, f+1-1 = f
     But for any smaller number (such as f/2) this is not true.  */

  sprintf (zBuf, "%.*eF", FLT_DIG + 2, FLT_EPSILON);
  sscanf (zBuf, "%e", &f);
  g = f + one;
  h = f / two + one;
  x ();
  if (g - one != f || h - one != 0.0)
    regress (zBuf, STRING(FLT_EPSILON));
  else
    ++test;

  /* Test the minimum normalized value.

     For a normalized value the full number of mantissa bits is
     available, so multiplying by 1+FLT_EPSILON sets the least
     significant bit of the mantissa.
     But for a denormalized value (such as f/2) we lose bits in
     the mantissa and multiplying by 1+FLT_EPSILON gives the
     number itself.  */

  sprintf (zBuf, "%.*eF", FLT_DIG + 2, FLT_MIN);
  sscanf (zBuf, "%e", &f);
  g = f * (1 + FLT_EPSILON);
  h = f / two * (1 + FLT_EPSILON);
  x ();
  if (f == 0.0 || f == g || f / two != h)
    regress (zBuf, STRING(FLT_MIN));
  else
    ++test;

  /* Test the maximum value.

     Test that it is not infinite (f != g) and that f * 2.0
     is infinite (g == h).  */

  sprintf (zBuf, "%.*eF", FLT_DIG + 2, FLT_MAX);
  sscanf (zBuf, "%e", &f);
  g = f * two;
  h = f * four;
  x ();
  if (f == 0.0 || g == f || g != h)
    regress (zBuf, STRING(FLT_MAX));
  else
    ++test;
}

void
test6_8 ()
{
  double f,g,h;
  double one = 1.0;
  double two = 2.0;
  double four = 4.0;
  char zBuf[ bufSize ];

  sprintf (zBuf, "%.*e", DBL_DIG + 2, DBL_EPSILON);
  sscanf (zBuf, "%le", &f);

  g = f + one;
  h = f / two + one;
  x ();
  if (g - one != f || h - one != 0.0)
    regress (zBuf, STRING(DBL_EPSILON));
  else
    ++test;

  sprintf (zBuf, "%.*e", DBL_DIG + 2, DBL_MIN);
  sscanf (zBuf, "%le", &f);
  g = f * (1 + DBL_EPSILON);
  h = f / two * (1 + DBL_EPSILON);
  x ();
  if (f == 0.0 || f == g || f / two != h)
    regress (zBuf, STRING(DBL_MIN));
  else
    ++test;

  sprintf (zBuf, "%.*e", DBL_DIG + 2, DBL_MAX);
  sscanf (zBuf, "%le", &f);
  g = f * two;
  h = f * four;
  x ();
  if (f == 0.0 || g == f || g != h)
    regress (zBuf, STRING(DBL_MAX));
  else
    ++test;
}

void
test9_11 ()
{
#ifdef HAVE_LONG_DOUBLE
  long double f,g,h;
  long double one = 1.0;
  long double two = 2.0;
  long double four = 4.0;
  char zBuf[ bufSize ];

  sprintf (zBuf, "%.*LeL", LDBL_DIG + 2, LDBL_EPSILON);
  sscanf (zBuf, "%Le", &f);

  g = f + one;
  h = f / two + one;
  x ();
  if (g - one != f || h - one != 0.0)
    regress (zBuf, STRING(LDBL_EPSILON));
  else
    ++test;

  sprintf (zBuf, "%.*LeL", LDBL_DIG + 2, LDBL_MIN);
  sscanf (zBuf, "%Le", &f);
  g = f * (1 + LDBL_EPSILON);
  h = f / two * (1 + LDBL_EPSILON);
  x ();
  if (f == 0.0 || f == g || f / two != h)
    regress (zBuf, STRING(LDBL_MIN));
  else
    ++test;

  sprintf (zBuf, "%.*LeL", LDBL_DIG + 2, LDBL_MAX);
  sscanf (zBuf, "%Le", &f);
  g = f * two;
  h = f * four;
  x ();
  if (f == 0.0 || g == f || g != h)
    regress (zBuf, STRING(LDBL_MAX));
  else
    ++test;
#else
  test += 3;
#endif
}

int
main( int argc, char** argv )
{
  test1 ();
  test2 ();
  test3_5 ();
  test6_8 ();

#if 0
  /* This is known to fail.  */
  test9_11 ();
#endif

  exit (wrong ? EXIT_FAILURE : EXIT_SUCCESS);
}

